/* program PL0
kompilator jazyka PL0 do postfixove notace a interpret */

#define NSYM 35		         /* pocet rozponatelnych symbol u*/
#define NORW 11              /* pocet klicovych slov */
#define TMAX 100             /* velikost tabulky identifikatoru */
#define AL 10                /* delka identifikatoru */
#define NMAX 5               /* maximalni pocet cislic v cisle */
#define CHSETSIZE 128        /* pocet znaku v mnozine */
#define MAXERR 30            /* maximalni pocet chyb */
#define LEVMAX 3             /* maximalni hloubka vnoreni */
#define AMAX 2048            /* nejvyssi adresa */
#define CXMAX 200            /* velikost prostoru pro kod */
#define STACKSIZE 500

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef enum {null, ident, number, plus, minus, times, slash, modulo, oddsym, eql,
              neq, lss, leq, gtr, geq, lparen, rparen, comma, semicolon, period,
              becomes, beginsym, endsym, ifsym, thensym, whilesym, dosym, callsym,
              constsym, varsym, procsym} SYMBOL;
typedef char ALFA[AL];
typedef enum {constant, variable, procedure} OOBJECT;
typedef int SYMSET[NSYM];
typedef enum {lit, opr, lod, sto, cal, ret, ing, jmp, jpc} FCT; /*funkce*/
/*
  lit 0,A        :uloz konstantu A do zasobniku
  opr 0,A        :proved instrukci A
  lod L,A        :uloz promenne L, A na vrchol zasobniku
  sto L,A        :zapis promennou L z vrcholu zasobniku do pameti
  cal L,A        :volej proceduru A z urovne L
  int 0,A        :zvys obsah T-registru o hodnotu A
  jmp 0,A        :proved skok na adresu A
  jpc 0,A        :proved podmineny skok na adresu A
*/

/* pridana polozka 'zeroindex' z duvodu generovani instrukce OPR
   s parametry od 1 */
typedef enum {zeroindex, neg,add,sub,mul,di,mod,odd,eq,ne,lt,ge,gt,le} Operation;
/*
  operace :
  neg	:negace hodnoty na vrcholu zasobniku
  add	:soucet dvou hodnot na vrcholu zasobniku
  sub	:rozdil dvou hodnot na vrcholu zasobniku
  mul	:soucin dvou hodnot na vrcholu zasobniku
  di	:deleni dvou hodnot na vrcholu zasobniku
  mod	:modulo dvou hodnot na vrcholu zasobniku
  odd	:test lichosti hodnot na vrcholu zasobniku
  eq	:test rovnost dvou hodnot na vrcholu zasobniku
  ne	:test nerovnost dvou hodnot na vrcholu zasobniku
  lt    :mensi
  ge	:vetsi nebo rovno
  gt	:vetsi
  le	:mensi nebo rovno
*/

typedef struct {
    FCT f;    /*kod funkce*/
    int l;    /*uroven*/
    int a;    /*cast adresy nebo kod operace*/
} INSTRUCTION;

char ch;                   /*posledni precteny znak*/
FILE *iva, *zdroj;          /*pomocny soubor pro vypis generovaneho kodu*/
                           /*a tabulky symbolu*/
int txpom;                 /*pomocna promenna*/
SYMBOL sym;                /*posledni precteny symbol*/
ALFA id;                   /*posledni precteny identifikator*/
int num;                   /*posledni prectene cislo*/
int cc;                    /*pocet znaku*/
int ll;                    /*delka radku*/
int kk,err;
int cx;                    /*pocitadlo adres*/
char line[81];             /*nacteny radek*/
ALFA a;
INSTRUCTION code[CXMAX+1]; /*pole generovanych kodu*/

ALFA word[NORW]={"begin", "call", "const", "do", "end", "if", "odd", "procedure",
		         "then", "var", "while"};

SYMBOL wsym[NORW]={beginsym, callsym, constsym, dosym, endsym, ifsym, oddsym,
		           procsym, thensym, varsym, whilesym};
SYMBOL ssym[255];

char mnemonic[][6]={"LIT  ", "OPR  ", "LOD  ", "STO  ", "CAL  ", "RET  ", "INT  ", "JMP  ", "JMC  "};

SYMSET declbegsys, statbegsys, facbegsys;

struct {
    ALFA name;
    OOBJECT kind;
    union {
        int val;
        struct {
            int level,adr,size;
        } vp;
    } CO;
} TABLE[TMAX+1];

int cx0,b,s[STACKSIZE];

void expression(SYMSET fsys,int tx,int lev);

/*vynulovat mnozinu symbolu co*/
void nuluj(SYMSET co) {
int i;

    for (i=0; i<30; i++) co[i]=0;
}   /* nuluj() */


/*sjednotit dve mnoziny symbolu*/
void sjednot(SYMSET s1,SYMSET s2) {
int i;

    for (i=0;i<30;i++) s1[i]=s1[i]||s2[i];
}   /* sjednot() */


void listtabsym(void) { /*pomocna fce pro vypis kon.tvaru tabulky symbolu*/
int i;

    fprintf(iva,"\n\ntabulka symbolu:\n");
    fprintf(iva,"***************\n");

    for (i=1;i<=txpom;i++) {
        fprintf(iva,"\n");
        fprintf(iva,"%4d  name: %-12s",i,TABLE[i].name);
    
        switch (TABLE[i].kind) {
            case constant: fprintf(iva," const val= %4d",TABLE[i].CO.val);
                           break;

            case variable: fprintf(iva," var   lev= %4d  adr= %4d  size= %4d",TABLE[i].CO.vp.level,TABLE[i].CO.vp.adr,TABLE[i].CO.vp.size);
		                   break;

            case procedure: fprintf(iva," proc  lev= %4d  adr= %4d  size= %4d",TABLE[i].CO.vp.level,TABLE[i].CO.vp.adr,TABLE[i].CO.vp.size);
                            break;

        }
    }
}   /* listtabsym() */


void listgencode(void) { /*pomocna fce pro vypis generovanych kodu*/
int i;

    fprintf(iva,"\ngenerovany kod:\n");
    fprintf(iva,"**************\n");

    for (i=0;i<=cx-1;i++)
        fprintf(iva,"\n%4d %.5s %3d %5d",i,mnemonic[code[i].f],code[i].l,code[i].a);

}   /* listgencode()	*/


void error(int n) {
int i;

    printf("****");
    for (i=0;i<cc-1;i++) printf(" ");
    printf("-%2d\n",n);
    if (++err>MAXERR) exit(1);

} /* error() */


/* nacita ze vstupniho souboru 1 znak do glob. promenne 'ch' a prekroci konec radky */
void getch(void) {

    if (cc == ll) {
        if (feof(zdroj)) {
            printf("program incompleted");
            exit(2);
        }
        ll = cc = 0;
        printf("       ");

        do {
            fscanf(zdroj,"%c",&ch);
            if ((ch != '\n') && (ch != '\r')) {
                line[ll++]=ch;
                printf("%c",ch);
            }
        } while ((ch != '\n') && (ch != '\r') && (feof(zdroj) == 0));

        printf("\n");
        line[ll++] = ' ';
    }
    ch = line[cc++];
}   /* getch()	*/


/* nacita ze vstupniho souboru 1 symbol a vrati jeho kod do glob. promenne 'sym' */
void getsym(void) {
int i, j, k;

    while (ch <= ' ') getch();
    if ((ch >= 'a') && (ch <= 'z')) {   /* identifier or reserved word */
        k = 0;

        do {
            if (k < AL) a[k++] = ch;
            getch();
        } while (((ch >= 'a') && (ch <= 'z')) || ((ch >= '0') && (ch <= '9')));

        a[k] = '\0';
        strcpy(id, a);
        i = 0;
        j = NORW - 1;

        do {
            k = (i + j) / 2;
            if (strcmp(id, word[k]) <= 0) j = k - 1;
            if (strcmp(id, word[k]) >= 0) i = k + 1;
        } while (i <= j);

        if ((i - 1) > j) sym = wsym[k];
        else sym = ident;
    }
    else
        if ((ch >= '0') && (ch <= '9')) {   /* number */
            k = num = 0;
            sym = number;

            do {
                num = 10 * num + (ch - '0');
                k++;
                getch();
            } while ((ch >= '0') && (ch <= '9'));
            if (k > NMAX) error(30);
        }
        else
            if (ch == ':') {
                getch();
                if (ch == '=') {
                    sym = becomes;
                    getch();
                }
                else sym = null;
            }
            else
                if (ch == '<') {
                    getch();
                    if (ch == '=') {
                        sym = leq;
                        getch();
                    }
                    else
                        if (ch == '>') {
                            sym = neq;
                            getch();
                        }
                        else
                            sym = lss;
                }
                else
                    if (ch == '>') {
                        getch();
                        if (ch == '=') {
                            sym = geq;
                            getch();
                        }
                        else sym = gtr;
                    }
                    else {
                        sym = ssym[ch];
                        getch();
                    }
} /* konec procedury getsym */


/* generovat instrukci do pole generovanych kodu
       x :kod funkce
       y :uroven instrukce
       z :adresni cast v instrukci
*/
void gen(FCT x,int y,int z) {

    if (cx>CXMAX) {
        printf("program too long");
        exit(1);
    }
    code[cx].f = x;
    code[cx].l = y;
    code[cx++].a = z;
}   /* gen()	*/


/* testovat zda nacteny symbol je v mnozine symbolu 's1'.
   Pokud neni generuje chybu a nacita opakovane ze vstupu
   dokud neni nacten pristupny symbol v mnozina 's1' a 's2'
*/
void test(SYMSET s1,SYMSET s2,int n) {
SYMSET pom;

    if (!s1[sym]) {
        error(n);
        nuluj(pom);
        sjednot(pom,s1);
        sjednot(pom,s2);

        while (pom[sym] == 0) getsym();
    }
}   /* test() */


/* vlozi object do tabulky symbolu
       k	:typ objektu, tj. zda jde o konstantu,promennou,...
       lev	:uroven, ve ktere je objekt deklarovan
       dx	:adresa objektu
*/
void enter(OOBJECT k,int *tx,int lev,int *dx) {

    (*tx)++;
    txpom = *tx; /* OR -- bast aby fungovala TS */
    strcpy(TABLE[*tx].name,id);
    TABLE[*tx].kind=k;

    switch (k) {
        case constant: if (num>AMAX) {
		                   error(31);
		                   num = 0;
                       }
                       TABLE[*tx].CO.val = num;
                       break;

        case variable: TABLE[*tx].CO.vp.level = lev;
		               TABLE[*tx].CO.vp.adr = (*dx)++;
		               break;

        case procedure: TABLE[*tx].CO.vp.level = lev;
		                break;

    }
}   /* enter() */


/* vyhleda symbol v tabulce symbolu
       id	:jmeno symbolu
       tx	:ukazovatko na konec tabulky symbolu
   navratova hodnota:
       -1 : symbol nenalezen
      >=0 : adresa symbolu
*/
int position(ALFA id,int tx) {
int i;

    strcpy(TABLE[0].name,id);
    i = tx;
    while (strcmp(TABLE[i].name,id)) i--;

return(i);
}   /* position() */


/* zpracova deklarace konstanty ve tvaru:
       ident = hodnota.
       tx	:ukazovatko na volne misto v tabulce symbolu
       lev	:uroven, ve ktere je symbol deklarovan
       dx	:adresa - neni pouzita, protoze jde tady o kontantu
*/
void constdeclaration(int *tx,int lev,int *dx) {

if (sym == ident) {
    getsym();
    if ((sym == eql) || (sym == becomes)) {
        if (sym == becomes) error(1);
        getsym();
        if (sym == number) {
            enter(constant,tx,lev,dx);
            getsym();
        } else error(2);
    } else error(3);
} else error(4);

}   /* constdeclaration() */


/* zpracova deklarace promenne
       tx	:ukazovatko na volne misto v tabulce symbolu
       lev	:uroven,ve ktere je symbol deklarovan
       dx	:adresa promenne
*/
void vardeclaration(int *tx,int lev,int *dx) {

if (sym == ident) {
    enter(variable,tx,lev,dx);
    getsym();
} else error(4);

}   /* vardeclaration() */


/*vypis generovanych kodu prelozeneho bloku*/
void listcode(void) {
int i;

    for (i=cx0;i<=(cx-1);i++)
        printf("%d %.5s %3d %5d\n",i,mnemonic[code[i].f],code[i].l,code[i].a);
}   /* listcode() */


/* generovat kod pro ulozeni hodnoty jednoho faktoru na vrcholu zasobniku,
   tj. bud promenna,kontanta,cislo nebo leva zavorka
       fsys	:mnozina dostupnych symbolu
	   (tj. faktor muze byt pouze promenna,kontanta,cislo nebo leva zavorka)

       tx	:ukazovatko na konec tabulky symbolu
       lev	:uroven faktoru
*/
void factor(SYMSET fsys,int tx,int lev) {
int i;
SYMSET pom;

    test(facbegsys,fsys,24);
    while (facbegsys[sym]) {
        if (sym == ident) {
            i = position(id,tx);
            if (i == 0) error(1);
            else
                switch (TABLE[i].kind) {
	                case constant: gen(lit,0,TABLE[i].CO.val);
                                   break;

	                case variable: gen(lod,lev-TABLE[i].CO.vp.level,TABLE[i].CO.vp.adr);
                                   break;

	                case procedure: error(21);
                                    break;

	            }
            getsym();
        } else
            if (sym == number) {
                if (num > AMAX) {
                    error(31);
                    num = 0;
                }
                gen(lit,0,num);
                getsym();
            } else
                if (sym == lparen) {
                    getsym();
                    nuluj(pom);
                    sjednot(pom,fsys);
                    pom[rparen] = 1;
                    expression(pom,tx,lev);
                    if (sym == rparen) getsym();
                    else error(22);
                }
                nuluj(pom);
                pom[lparen] = 1;
                test(fsys,pom,23);
    }
}   /* factor() */


/* generovat kody pro vypocet jednoho termu,
   tj. pro vyhodnoceni vyrazu ve tvaru faktor*faktor,faktor/faktor nebo faktor%faktor
       fsys	:mnozina dostupnych symbolu ve termu
       tx	:ukazovatko na konec tabulky symbolu
       lev	:uroven termu
*/
void term(SYMSET fsys,int tx,int lev) {
SYMBOL mulop;
SYMSET pom;

    nuluj(pom);
    sjednot(pom,fsys);
    pom[times] = pom[slash] = pom[modulo] = 1;
    factor(pom,tx,lev);

    while ((sym == times) || (sym == slash) || (sym == modulo)) {
        mulop = sym;
        getsym();
        sjednot(pom,fsys);
        pom[times] = pom[slash] = pom[modulo] = 1;
        factor(pom,tx,lev);

        switch(mulop) {
            case times: gen(opr,0,mul);
                        break;

            case slash: gen(opr,0,di);
                        break;

            case modulo: gen(opr,0,mod);
                         break;

        }
    }

}   /* term() */


/* generovat kody pro vyhodnoceni aritm. vyrazu
       fsys	:mnozina dostupnych symbolu ve vyrazu
       tx	:ukazovatko na konec tabulky symbolu
       lev	:uroven vyrazu
*/
void expression(SYMSET fsys,int tx,int lev) {
SYMBOL addop;
SYMSET pom;

    if ((sym == plus) || (sym == minus)) {
        addop = sym;
        getsym();
        nuluj(pom);
        sjednot(pom,fsys);
        pom[plus] = pom[minus] = 1;
        term(pom,tx,lev);
        if (addop == minus) gen(opr,0,neg);
    }
    else {
        nuluj(pom);
        sjednot(pom,fsys);
        pom[plus] = pom[minus] = 1;
        term(pom,tx,lev);
    }

    while ((sym == plus) || (sym == minus)) {
        addop = sym;
        getsym();
        nuluj(pom);
        sjednot(pom,fsys);
        pom[plus] = pom[minus] = 1;
        term(pom,tx,lev);

        if (addop == plus) gen(opr,0,add);
        else gen(opr,0,sub);
    }
}   /* expression() */


/* generovat kody pro vyhodnoceni logickeho vyrazu
       fsys	:dostupne symoboly v vyrazu
       tx	:ukazovatko na konec tabulky symbolu
       lev	:uroven, ve ktere je prelozeny vyraz
*/
void condition(SYMSET fsys,int tx,int lev) {
SYMBOL relop;
SYMSET pom;

    if (sym == oddsym) {
        getsym();
        expression(fsys,tx,lev);
        gen(opr,0,odd);
    }
    else {
        nuluj(pom);
        sjednot(pom,fsys);
        pom[eql] = pom[neq] = pom[lss] = pom[gtr] = pom[leq] = pom[geq] = 1;
        expression(pom,tx,lev);
        if ((sym != eql) && (sym != neq) && (sym != lss) && (sym != gtr) && (sym != leq) && (sym != geq))
            error(22);
        else {
            relop = sym;
            getsym();
            expression(fsys,tx,lev);

            switch (relop) {
                case eql: gen(opr,0,eq);
                          break;

                case neq: gen(opr,0,ne);
                          break;

                case lss: gen(opr,0,lt);
                          break;

                case geq: gen(opr,0,ge);
                          break;

                case gtr: gen(opr,0,gt);
                          break;

                case leq: gen(opr,0,le);
                          break;

            }
        }
    }

}   /* condition() */


/* generovat kody pro statement
       fsys	:dostupne symboly v statmentu
       tx	:ukazovatko na konec tabulky symbolu
       lev	:uroven, ve ktere je prelozeny statement
*/
void statement(SYMSET fsys,int tx,int lev) {
int i, cx1, cx2;
SYMSET pom;

    if ((fsys[sym] == 0) && (sym != ident)) {
        error(10);
        do
            getsym();
        while (fsys[sym] == 0);
    }
    if (sym == ident) { /*nalezen prikaz prirazeni*/
        i = position(id,tx);
        if (i == 0) error(11);
        else
            if (TABLE[i].kind!=variable) { /*prirazeni do jineho ident. nez promenna*/
                error(12);
                i = 0;
            }
        getsym();
        if (sym == becomes) getsym();
        else error(13);
        expression(fsys,tx,lev);
        if (i) gen(sto,lev-TABLE[i].CO.vp.level,TABLE[i].CO.vp.adr);
    }
    else
        if (sym == callsym) {/*nalezeno volani podprogramu*/
            getsym();
            if (sym != ident) error(14);
	        else {
	            if ((i = position(id,tx)) == 0) error(11);
	            else {
	                if (TABLE[i].kind == procedure) gen(cal,lev-TABLE[i].CO.vp.level,TABLE[i].CO.vp.adr);
					else error(15);
                }
	            getsym();
	        }
        }
        else
            if (sym == ifsym) { /*podmineny prikaz*/
                getsym();
                nuluj(pom);
                sjednot(pom,fsys);
                pom[thensym] = pom[dosym] = 1;
                condition(pom,tx,lev);

                if (sym == thensym) getsym();
                else error(16);
                cx1 = cx;
                gen(jpc,0,0);
                statement(fsys,tx,lev);
                code[cx1].a = cx;
            }
            else
                if (sym == beginsym) { /*zacina novy blok*/
                    getsym();
                    nuluj(pom);
                    sjednot(pom,fsys);
                    pom[semicolon] = pom[endsym] = 1;
                    statement(pom,tx,lev);

                    while ((statbegsys[sym]) || (sym == semicolon)) {
                        if (sym == semicolon) getsym();
                        else error(10);
                        sjednot(pom,fsys);
                        pom[semicolon] = pom[endsym] = 1;
                        statement(pom,tx,lev);
                    }

                    if (sym == endsym) getsym(); /*konci predchozi blok*/
                    else error(17);
                }
                else
                    if (sym == whilesym) { /*zacina cyklus while*/
                        cx1 = cx;
                        getsym();
                        nuluj(pom);
                        sjednot(pom,fsys);
                        pom[dosym] = 1;
                        condition(pom,tx,lev);
                        cx2 = cx;
                        gen(jpc,0,0);
                        if (sym == dosym) getsym();
                        else error(18);
                        statement(fsys,tx,lev);
                        gen(jmp,0,cx1);
                        code[cx2].a = cx;
                    }

    test(fsys,pom,19); /*test koreknost ukonceni statementu*/

}   /* statement() */


/* generovat kody pro blok ve tvaru
   const a,b,c,..;
   var x,y,z,..;
   podprogramy
   begin
   ...
   end.

       lev	:uroven bloku
       tx	:ukazovatko na konec tabulky symbolu
       fsys	:dostupne symoboly v bloku
*/
void block(int lev,int tx,SYMSET fsys) {
int tx0, dx;
SYMSET pom;

    dx = 3;
    tx0 = tx;
    TABLE[tx].CO.vp.adr = cx;
    gen(jmp,0,0);
    if (lev > LEVMAX) error(32);

    do {
        if (sym == constsym) { /*deklaracni cast konstant*/
            getsym();

            do {
                constdeclaration(&tx,lev,&dx);
                while (sym == comma) {
                    getsym();
                    constdeclaration(&tx,lev,&dx);
                }
                if (sym == semicolon) getsym();
                else error(5);
            } while (sym==ident);
        }
        if (sym == varsym) { /*deklaracni cast promennych*/
            getsym();
            vardeclaration(&tx,lev,&dx);

            while (sym == comma) {
                getsym();
                vardeclaration(&tx,lev,&dx);
            }
            if (sym == semicolon) getsym();
		    else error(5);
        }
        while (sym == procsym) { /*definice podprogramu*/
            getsym();
            if (sym == ident) {
                enter(procedure,&tx,lev,&dx);
                getsym();
            } else error(4);
            if (sym==semicolon) getsym();
            else error(5);
            nuluj(pom);
            sjednot(pom,fsys);
            pom[semicolon] = 1;
            block(lev+1,tx,pom);
            if (sym == semicolon) {
                getsym();
                nuluj(pom);
                sjednot(pom,statbegsys);
                pom[ident] = pom[procsym] = 1;
                test(pom,fsys,6);
            } else error(5);
        }
        nuluj(pom);
        sjednot(pom,statbegsys);
        pom[ident] = 1;
        test(pom,declbegsys,7);
    } while (declbegsys[sym]);

    code[TABLE[tx0].CO.vp.adr].a = cx;
    TABLE[tx0].CO.vp.adr = cx;
    TABLE[tx0].CO.vp.size = dx;
    cx0 = cx;
    gen(ing,0,dx);
    nuluj(pom);
    sjednot(pom,fsys);
    pom[semicolon] = pom[endsym] = 1;
    statement(pom,tx,lev);
    gen(ret,0,0);
    test(fsys,pom,8);
    listcode();

}   /* block() */


/* funkce pro zjisteni pocatecni adresy na urovni, ktera je o l vyssi 
   nez soucasna uroven, ve ktere je prave vykonan program
*/
int base(int l) {
int b1;

    b1 = b;
    while (l > 0) {
        b1 = s[b1];
        l--;
    }

return(b1);
}

/* interpretace generovanych kodu */
void interpret(void) {
int p, t;
INSTRUCTION i;

    printf("START PL/0\n");
    t = p = s[1] = s[2] = s[3] = 0;
    b = 1;
    do {
        i = code[p++];

        switch (i.f) {
            case lit: s[++t] = i.a;
                      break;

            case opr:
                switch (i.a) {
                    case neg : s[t] = -s[t];
			                   /*printf("INTERPRET OPR: neg\n");*/
			                   break;

                    case add : t--;
                               s[t] += s[t + 1];
			                   /*printf("INTERPRET OPR: add\n");*/
			                   break;

	                case sub : t--;
			                   s[t] -= s[t + 1];
			                   /*printf("INTERPRET OPR: sub\n");*/
			                   break;

                    case mul : t--;
			                   s[t] *= s[t + 1];
			                   /*printf("INTERPRET OPR: mul\n");*/
			                   break;

	                case di :  t--;
			                   if (s[t + 1] != 0) s[t] /= s[t + 1];
                               else {
				   /* error */
                                   error(31);
                               }
                               /*printf("INTERPRET OPR: di\n");*/
                               break;

	                case mod : t--;
			                   s[t] = s[t] % s[t + 1];
			                   /*printf("INTERPRET OPR: mod\n");*/
			                   break;

                    case odd : s[t] = s[t] % 2;
			                   /*printf("INTERPRET OPR: odd\n");*/
			                   break;

                    case eq : t--;
		                      s[t] = (s[t] == s[t + 1]);
			                  /*printf("INTERPRET OPR: eq\n");*/
                              break;

	                case ne : t--;
		                      s[t] = (s[t] != s[t + 1]);
			                  /*printf("INTERPRET OPR: ne\n");*/
                              break;

	                case lt : t--;
		                      s[t] = (s[t] < s[t + 1]);
			                  /*printf("INTERPRET OPR: lt\n");*/
                              break;

	                case ge : t--;
		                      s[t] = (s[t] >= s[t + 1]);
			                  /*printf("INTERPRET OPR: ge\n");*/
		                      break;

                    case gt : t--;
		                      s[t] = (s[t] > s[t + 1]);
			                  /*printf("INTERPRET OPR: gt\n");*/
                              break;

	                case le : t--;
		                      s[t] = (s[t] <= s[t + 1]);
			                  /*printf("INTERPRET OPR: le\n");*/
		                      break;

	               }
	               break;

            case lod: t++;
	                  s[t] = s[base(i.l) + i.a];
	                  break;

            case sto: s[base(i.l) + i.a] = s[t];
	                  printf("%d\n",s[t--]);
	                  break;

            case cal: s[t + 1] = base(i.l);
	                  s[t + 2] = b;
	                  s[t + 3] = p;
	                  b = t + 1;
	                  p = i.a;
                      break;

            case ret: t = b - 1;
	                  p = s[t + 3];
	                  b = s[t + 2];
	                  break;

            case ing: t += i.a;
                      break;

            case jmp: p = i.a;
                      break;

            case jpc: if (s[t] == 0) p = i.a;
	                  t--;
	                  break;

        }
    } while (p);
    printf(" END PL/0\n");

}   /* interpret() */


/*hlavni program*/
main(void) {
SYMSET pom;
char zdrojak[13];

    /* cte jmeno souboru, dokud uzivatel nezada nenulovy retezec */
    do {
        printf("Zadej jmeno souboru obsahujiciho zdrojovy text: ");
        scanf("%s",zdrojak);
    } while (strlen(zdrojak) < 1);

    if ((iva = fopen("TAB.SYM", "w")) == NULL) {
        printf("\nCHYBA! Nepodarilo se otevrit soubor pro zapis tabulky symbolu...\n");
        return(-1);
    }

    /* ...a pak otestuje, jestli soubor existuje */
    if ((zdroj = fopen(zdrojak, "r")) == NULL) {
        printf("\nCHYBA! Nepodarilo se otevrit soubor se zdrojovym textem [%s]...\n",zdrojak);
        return(-1);
    }

    for (ch=' ';ch<='_';ch++) ssym[ch] = null;
    ssym['+'] = plus;
    ssym['-'] = minus;
    ssym['*'] = times;
    ssym['/'] = slash;
    ssym['%'] = modulo;
    ssym['('] = lparen;
    ssym[')'] = rparen;
    ssym['='] = eql;
    ssym[','] = comma;
    ssym['.'] = period;
    ssym['#'] = neq;
    ssym['<'] = lss;
    ssym['>'] = gtr;
    ssym[';'] = semicolon;
    nuluj(declbegsys);
    nuluj(statbegsys);
    nuluj(facbegsys);

    /*v deklaracni casti se musi zacinat bud 'const','var' nebo 'procedure'*/
    declbegsys[constsym] = declbegsys[varsym] = declbegsys[procsym] = 1;

    /*ve statementu se musi zacinat bud 'begin','call','if','while' nebo ident.*/
    statbegsys[beginsym] = statbegsys[callsym] = statbegsys[ifsym] = statbegsys[whilesym] = 1;

    /*faktor muze byt bud ident., cislo nebo leva zavorka*/
    facbegsys[ident] = facbegsys[number] = facbegsys[lparen] = 1;

    err = cc = cx = ll = 0;
    ch = ' ';
    kk = AL;
    getsym();
    nuluj(pom);
    sjednot(pom,declbegsys); /*program zacina deklaracni casti*/
    sjednot(pom,statbegsys); /*nasleduje cast kodu hlavniho programu*/
    pom[period] = 1;
    block(0,0,pom); /*zavola preklad programu*/
    if (sym != period) error(9); /*a konci teckou*/
    listgencode();
    listtabsym();
    if (err == 0) {
        printf("\nno error in PL/0 program\n");
        interpret(); /*interpretuje generovane kody, nejsou-li chyby v programu*/
    }
    else printf("\n %2d error(s) in PL/0 program",err);

return(0);
}
